/************************************************************************************************************************/
/*

Software for smart cameras from Vision Components
-------------------------------------------------

Program: Klaus Schneider, VC

Data Matrix Reader


Program Description:
--------------------

Version 23:
-----------
- scale independent
- accurate bounding box
- speed improvements
- easy setup parameters [0..9]
- cache optimization
- flexible anchor point
- generate xy list for pattern position

Version 24:
-----------
- scale mode: detecting pattern with better correlation values first, independent of pattern size

Version 25:
-----------
- 1000 runs and 90 min time limit in demo mode
- include lib version in PatPar struct (PatPar->LibVersion)
*/
/************************************************************************************************************************/

#include "SmartFinder.h"

/************************************************************************************************************************/

#define TRANRED    1
#define TRANYEL    3
#define COLGREY    4
#define COLRED     8
#define COLBLACK  16
#define COLWHITE  32
#define COLGREEN  64
#define COLBLUE  128

/************************************************************************************************************************/

// global values

char CompName[] = "Vision Components";
char ProgName[] = "VC Smart Finder";
I32  ProgVer;

/************************************************************************************************************************/
/************************************************************************************************************************/
/************************************************************************************************************************/

/////////////////////////////////////////////////////////////////////////////////
// demo program
/////////////////////////////////////////////////////////////////////////////////

/************************************************************************************************************************/
/************************************************************************************************************************/
/************************************************************************************************************************/

/************************************************************************************************************************/

void main (void)
{
  char        response, OutputText[100];
  I32         AutoMode, VideoDisplCamera, sec, ms, Error;
  I32         ScreenX, ScreenY, ScreenDx, ScreenDy, SearchBorder = 8;
  I32         SearchAreaX, SearchAreaY, SearchAreaDx, SearchAreaDy, PatX, PatY, PatDx, PatDy;
  image       ImgOvl;
  image       SearchArea, Pattern;
  
  // matching variables
  PatternPar  PatPar[1];
  I32         TeachReady, NextFind;

  // text display parameter
  I32   PosX, PosY, WriteColor=COLBLUE;
  image ImgTxt;


  /////////////////////////////////////////////////////////////////////////////////
  // init all VC library licences
  /////////////////////////////////////////////////////////////////////////////////

  /*******************************************************************************/
  /*                 PLEASE INITIALIZE YOUR LICENCE CODES                        */
  /*                    To activate library functions                            */
  /*                                                                             */
  /* VCLIB :         init_licence("T1122334455") or init_licence("LT1122334455") */
  /* Extension Lib : init_licence("E1122334455") or init_licence("LE1122334455") */
  /* Color Lib :     init_licence("C1122334455") or init_licence("LC1122334455") */
  /* etc.                                                                        */
  /*******************************************************************************/

  // if "licence.h" is available
  INIT_LICENCE();


  /////////////////////////////////////////////////////////////////////////////////
  // activate VC Keypad inputs
  /////////////////////////////////////////////////////////////////////////////////

  INTERFACE_MODE(SERIAL);
  

  /////////////////////////////////////////////////////////////////////////////////
  // ini smart finder library memory and parameters (one time only)
  /////////////////////////////////////////////////////////////////////////////////

  print("Start %s. Please wait... \n", ProgName); 

  Error      = 0;
  if (Error == 0) Error = SetParameter(PatPar);
  if (Error == 0) Error = InitSmartFinder(PatPar);
  
  if (Error) {print("Error %d: Init Smart Finder\n", Error); goto END;}
  
  // read VC Smart Reader Library Version
  ProgVer = PatPar->LibVersion;

 

  /////////////////////////////////////////////////////////////////////////////////
  // ini variables
  /////////////////////////////////////////////////////////////////////////////////

  NextFind   = 0;
  AutoMode   = 0;
  TeachReady = 0;
  

  // ini display and define Screen Size and Positon (Screen = image capture position)
  IniCamera  (&ScreenX, &ScreenY, &ScreenDx, &ScreenDy);
  
  // define search area 
  SearchAreaDx       =  ScreenDx - 2 * SearchBorder;
  SearchAreaDy       =  ScreenDy - 2 * SearchBorder;
  SearchAreaX        =  ScreenX + (ScreenDx -  SearchAreaDx) / 2;
  SearchAreaY        =  ScreenY + (ScreenDy -  SearchAreaDy) / 2;

  ImageAssign(&SearchArea, ScrByteAddr(SearchAreaX, SearchAreaY), SearchAreaDx, SearchAreaDy, ScrGetPitch);  
  

  // define teach area 
  PatDx              =  SearchAreaDx / 4;
  PatDy              =  SearchAreaDy / 4;
  PatX               =  SearchAreaX + (SearchAreaDx - PatDx) / 2;
  PatY               =  SearchAreaY + (SearchAreaDy - PatDy) / 2;
       
  ImageAssign(&Pattern, ScrByteAddr(PatX, PatY), PatDx, PatDy, ScrGetPitch);  


  
  // system variable DispGetRows and DispGetColumns = 0 for cameras witout video output 
  if ( (DispGetRows == 0) ||  (DispGetColumns == 0) ) VideoDisplCamera = 0;
  else                                                VideoDisplCamera = 1; 



  /////////////////////////////////////////////////////////////////////////////////
  // start patterm matching
  /////////////////////////////////////////////////////////////////////////////////

  print("\n");
  print("--- %s Version %d.%d ---\n", ProgName, ProgVer/10, ProgVer%10);
  print("\n");

  print("Press ENTER to start single match\n");
  print("Press  F1   to teach pattern\n");
  print("Press  'n'  next match\n");
  print("Press  'p'  to change search area and parameter\n");
  print("Press  's'  to store pattern at fd:\n");
  print("Press  'r'  to read pattern from fd:\n");
  print("Press  'q'  to quit program\n");
  print("Press  'g'  to toggle automatic trigger mode\n");

  print("\n");

  do
  {
    Error = 0;
    
    // print("Mem=%x: \n", sysmemfree() );

    /////////////////////////////////////////////////////////////////////////////////
    // draw search area into overlay
    /////////////////////////////////////////////////////////////////////////////////

    ImageAssign(&ImgOvl,     OvlByteAddr(SearchAreaX, SearchAreaY), SearchAreaDx, SearchAreaDy, OvlGetPitch);  
    framed(&ImgOvl, COLBLUE);
    
    if(VideoDisplCamera) vmode(vmOvlLive);


    do
    {
      //////////////////////////////////////////////////////////////
	    // read new command
      //////////////////////////////////////////////////////////////

      response = KEY_INVALID;

      if (kbready() || rbready()) 
      {
        response = ReadKeyboard(1);
        EmptyKeyboardBufferAll();
	    }


      //////////////////////////////////////////////////////////////
      // teach pattern
      //////////////////////////////////////////////////////////////

      if (response == KEY_F1 || response == 't'  || TeachReady == 0) 
      {
        if(VideoDisplCamera) set(&ImgOvl,0);  
        framed(&ImgOvl, COLBLUE);

        // define search area overlay
        print("\nPlease select pattern area\n"); 
        SelectArea(&PatX, &PatY, &PatDx, &PatDy, SearchAreaX, SearchAreaX+SearchAreaDx, SearchAreaY, SearchAreaY+SearchAreaDy, 0, 0, COLBLUE);

        ImageAssign(&Pattern, ScrByteAddr(PatX, PatY), PatDx, PatDy, ScrGetPitch);  

        Error = TeachPattern(PatPar, &Pattern);
        
        TeachReady = 1;
      }

      //////////////////////////////////////////////////////////////
      // change parameter
      //////////////////////////////////////////////////////////////

      if (response == 'p') 
      {
        // define search area overlay
        print("\nPlease select search area\n"); 
        SelectArea(&SearchAreaX, &SearchAreaY, &SearchAreaDx, &SearchAreaDy, ScreenX, ScreenX+ScreenDx, ScreenY, ScreenY+ScreenDy, 0, 0, COLBLUE);

        ChangeParameter (PatPar);

        ImageAssign(&SearchArea, ScrByteAddr(SearchAreaX, SearchAreaY), SearchAreaDx, SearchAreaDy, ScrGetPitch);  
      }

      
      //////////////////////////////////////////////////////////////
      // search for next pattern / different pattern
      //////////////////////////////////////////////////////////////

      if (response ==    'n'    ) {NextFind=1; goto START;}
      if (response == KEY_ENTER ) {NextFind=0; goto START;}


      //////////////////////////////////////////////////////////////
      // store and read pattern from file
      //////////////////////////////////////////////////////////////

      if (response ==   's'  ) StorePattern(PatPar, "fd:/Pattern.dat");
      if (response ==   'r'  ) ReadPattern (PatPar, "fd:/Pattern.dat");



      if (response == KEY_F2 ) ChangeParameterTest(PatPar);

      // if (response == KEY_ESC) goto END;
      if (response ==   'q'  ) goto END;

      if (response ==   'g'  ) if(AutoMode) AutoMode=0; else AutoMode=1; 
    }
    while( AutoMode == 0 );
    
    
    
    /////////////////////////////////////////////////////////////////////////////////
    // start find pattern
    /////////////////////////////////////////////////////////////////////////////////

    START:
    
    /////////////////////////////////////////////////////////////////////////////////////
    // do pattern matching
    /////////////////////////////////////////////////////////////////////////////////////

    if (NextFind == 0) tpict();

    ms  = getvar(MSEC); sec = getvar(SEC);

    // now find pattern
    if (NextFind == 0) 
    {
      if (Error == 0) Error = FindPattern(PatPar, &SearchArea, FIND_PAT); 
    }
    else           
    {
      if (Error == 0) Error = FindPattern(PatPar, &SearchArea, DELETE_PAT); 
      if (Error == 0) Error = FindPattern(PatPar, &SearchArea,   NEXT_PAT); 
    }
    
    ms  = 1000 * (getvar(SEC) - sec) + (getvar(MSEC) - ms);


    /////////////////////////////////////////////////////////////////////////////////
    // delete old overlay drawings
    /////////////////////////////////////////////////////////////////////////////////

    ImageAssign(&ImgOvl, OvlByteAddr(ScreenX, ScreenY), ScreenDx, ScreenDy, OvlGetPitch); 
    if (NextFind == 0) if(VideoDisplCamera) set(&ImgOvl, 0); 


    if (Error == 0)
    {
      // draw bounding box
      ImageAssign(&ImgOvl, OvlByteAddr(SearchAreaX + PatPar->BoundingX, SearchAreaY + PatPar->BoundingY), PatPar->BoundingDx, PatPar->BoundingDy, OvlGetPitch);  
      // framed(&ImgOvl, COLRED);

      // draw pattern
      ImageAssign(&ImgOvl, OvlByteAddr(SearchAreaX, SearchAreaY), SearchAreaDx, SearchAreaDy, OvlGetPitch); 
      Error = DrawPattern(PatPar, &ImgOvl, COLGREEN);
      
      // draw anchor point
      ImageAssign(&ImgOvl, OvlByteAddr(SearchAreaX, SearchAreaY), SearchAreaDx, SearchAreaDy, OvlGetPitch);  
      MarkCross(&ImgOvl, Round(PatPar->PosX), Round(PatPar->PosY), COLRED, 2);
      
      // generate pattern list XY (I32 pointer list with 2 * PatPar->PatListLenXY entries [x0, y0, x1, y1, ...)
      /*{ 
        I32 i, *PatXY;

        PatXY = (I32 *)sysmalloc(2 * PatPar->PatListLenXY, MDATA);
        ListPattern(PatPar, PatXY);
        sysfree(PatXY);
      }*/

      // generate result string
      sprint(OutputText, "Error=%d: X=%6.1f  Y=%6.1f  Phi=%6.1f  Scale=%5.1f%%  Corr=%5.1f%%  Time=%3dms", Error, PatPar->PosX, PatPar->PosY, PatPar->Phi, 100.0 * PatPar->Scale, 100.0 * PatPar->Match, ms);
    }
    else
    {
      if (Error  > -2000) sprint(OutputText, "Error=%d:  Memory Error.", Error);
      if (Error == -2000) sprint(OutputText, "Error=%d:  Please verify VC licence code.", Error);
      if (Error == -2002) sprint(OutputText, "Error=%d:  No Pattern Edges.", Error);
      if (Error == -2003) sprint(OutputText, "Error=%d:  Search Area not ready. Call FIND_PAT first.", Error);
      if (Error == -2004) sprint(OutputText, "Error=%d:  Wrong Teach Image Format.", Error);
      if (Error == -2005) sprint(OutputText, "Error=%d:  InitSmartFinder() missing.", Error);
      if (Error == -2006) sprint(OutputText, "Error=%d:  No Pattern Tought.", Error);
      if (Error == -2007) sprint(OutputText, "Error=%d:  Pattern requests different Pyramid Zoom.", Error);
    }

    // display results string on video monitor
    PosX  = ScreenX;
    PosY  = ScreenY + ScreenDy + 6;

    PosY += 10;
    ImageAssign(&ImgTxt, OvlByteAddr(PosX, PosY), ScreenDx, 8, OvlGetPitch);  
    set(&ImgTxt, COLGREY);
    chprint1(OutputText, &ImgTxt, 1, 1, WriteColor);
    ChangeColor(&ImgTxt, 0, 0, COLGREY);

    #ifdef TESTING_VERSION
      print("%s\n", OutputText);
    #endif
  }
  while(response != KEY_ESC);

  
  /////////////////////////////////////////////////////////////////////////////////
  // end of VC Smart Reader program
  /////////////////////////////////////////////////////////////////////////////////

  END:

  DeinitSmartFinder(PatPar);
  
  IniCameraBack();
}

/*************************************************************************************************************************/

void IniCamera (I32 *ScreenX, I32 *ScreenY, I32 *ScreenDx, I32 *ScreenDy)

{
  I32   ScrX, ScrY, ScrDx, ScrDy;
  I32   VideoDisplCamera, FullScreenCamera, BorderDx, BorderDy;
  char  string[80];
  image Ovl, Area;


  /////////////////////////////
  // Ini Overlay Border
  /////////////////////////////

  BorderDx = 16;
  BorderDy = 64;


  /////////////////////////////
  // Define Camera Type
  /////////////////////////////

  // cameras witout video output have DispGetRows and DispGetColumns = 0
  if ( (DispGetRows == 0) ||  (DispGetColumns == 0) ) 
  {
    VideoDisplCamera = 0;
    FullScreenCamera = 1;
  }
  else
  {
    VideoDisplCamera = 1;

    // cameras sensor resolution compared with display resolution
    if (abs(DispGetRows - ScrGetRows) > BorderDy) 
    {
      FullScreenCamera = 0;
    }
    else
    {
      FullScreenCamera = 1;
    }
  }

  // print("\nVideoDisplCamera = %d\nFullScreenCamera = %d\n", VideoDisplCamera, FullScreenCamera);


  /////////////////////////////
  // Define Overlay Color
  /////////////////////////////

  if(VideoDisplCamera)
  {
    /* Define Color */
    set_translucent(1, 250,   0,   0); /* TRANRED    1 red translucent */
    set_translucent(2, 255, 255,   0); /* TRANYEL    3 yellow translucent */
    set_overlay_bit(2, 210, 210, 210); /* COLGREY    4 grey */
    set_overlay_bit(3, 250,   0,   0); /* COLRED     8 red */
    set_overlay_bit(4,  70,  70,  70); /* COLBLACK  16 black */
    set_overlay_bit(5, 250, 250, 250); /* COLWHITE  32 white */
    set_overlay_bit(6,   0, 200,   0); /* COLGREEN  64 green */
    set_overlay_bit(7,  65,  90, 255); /* COLBLUE  128 blue */
  }


  /////////////////////////////
  // Define Logical Pages
  /////////////////////////////

  // screen page
  ScrSetLogPage((I32)ScrGetPhysPage);

  // overlay page
  if(VideoDisplCamera)
  {
    // draw into overlay
    OvlSetLogPage((I32)OvlGetPhysPage);
  }
  else
  {
    // draw into screen
    OvlSetLogPage((I32)ScrGetPhysPage);
  }

  /////////////////////////////
  // Position of the image
  /////////////////////////////

   // Define result output border
   ScrDx = ScrGetColumns - 2 * BorderDx;
   ScrDy = ScrGetRows    - 2 * BorderDy;
   ScrX  = (ScrGetColumns-ScrDx)/2;
   ScrY  = (ScrGetRows   -ScrDy)/2;

  if (FullScreenCamera == 0) 
  {
    // Define screen position in the center. X-position of sensor image must be in allignment of 32
    ScrDx = ScrGetColumns;
    ScrDy = ScrGetRows;
    ScrX  = ((DispGetColumns-ScrDx)/64)*32;
    ScrY  = ((DispGetRows   -ScrDy)/2);

    CapturePosition(ScrX, ScrY);
  }


  /////////////////////////////
  // clear overlay and image
  /////////////////////////////

  if (VideoDisplCamera) 
  {
    ImageAssign(&Ovl,  getvar(OVLY_START), DispGetColumns, DispGetRows, OvlGetPitch);
    set(&Ovl,0);

    ImageAssign(&Area, getvar(DISP_START), DispGetColumns, DispGetRows, DispGetPitch);
    set(&Area, 210);
  }


  /////////////////////////////
  // Draw Screen and Overlay
  /////////////////////////////

  if (VideoDisplCamera) 
  {
    // set background color
    ImageAssign(&Ovl, OvlByteAddr(0, 0), DispGetColumns, DispGetRows, OvlGetPitch);
    set(&Ovl, COLGREY);

    // open screen size
    ImageAssign(&Ovl, OvlByteAddr(ScrX, ScrY), ScrDx, ScrDy, OvlGetPitch);
    set(&Ovl, 0);
    Low(&Ovl);
    
    // Company
    ImageAssign(&Ovl, OvlByteAddr(ScrX, 20), 400, 16, OvlGetPitch);
    chprint1(CompName,&Ovl, 2, 2, COLBLUE);
    ChangeColor(&Ovl,0,0,COLGREY);

    // Version
    sprint(string, "%s %d.%d", ProgName, ProgVer/10, ProgVer%10);
    ImageAssign(&Ovl, OvlBitAddr(ScrX + ScrDx - 8*strlen(string), 28), 8*strlen(string), 8, OvlGetPitch);
    chprint1(string, &Ovl, 1, 1, COLBLUE);
    ChangeColor(&Ovl,0,0,COLGREY);

    // live image with overlay
    set_ovlmask(255);
    vmode(vmOvlLive);

    if (FullScreenCamera == 0) 
    {
      // now refer to new screen position
      ScrSetLogPage((I32)ScrByteAddr(ScrX, ScrY));
      OvlSetLogPage((I32)OvlByteAddr(ScrX, ScrY));

      ScrX  = 0;
      ScrY  = 0;
      ScrDx = ScrGetColumns;
      ScrDy = ScrGetRows;
    }
  }

  *ScreenX  = ScrX;
  *ScreenY  = ScrY;
  *ScreenDx = ScrDx;
  *ScreenDy = ScrDy;
}

/*************************************************************************************************************************/

void IniCameraBack (void)

{
  I32   VideoDisplCamera;
  image Img;


  /////////////////////////////
  // Define Camera Type
  /////////////////////////////

  // cameras witout video output have DispGetRows and DispGetColumns = 0
  if ( (DispGetRows == 0) ||  (DispGetColumns == 0) ) 
  {
    VideoDisplCamera = 0;
  }
  else
  {
    VideoDisplCamera = 1;
  }


  /////////////////////////////
  // Position of the image
  /////////////////////////////

  if(VideoDisplCamera)
  {
    ScrSetPhysPage(getvar(DISP_START));
    ScrSetLogPage (getvar(DISP_START));
    OvlSetLogPage (getvar(OVLY_START));
  }

  /////////////////////////////
  // stop live mode
  /////////////////////////////

  tpict();

  /////////////////////////////
  // clear overlay and image
  /////////////////////////////

  if(VideoDisplCamera)
  {
    ImageAssign(&Img, OvlByteAddr(0, 0), DispGetColumns, DispGetRows, ScrGetPitch);
    set(&Img,0);

    ImageAssign(&Img, ScrByteAddr(0, 0), DispGetColumns, DispGetRows, ScrGetPitch);
    set(&Img,0);

    // refresh memory
    clean_cache();
  }


  /////////////////////////////
  // switch to live mode
  /////////////////////////////

  vmode(vmLive);
}

/*************************************************************************************************************************/

I32 SetParameter (PatternPar *PatPar)
{
  //////////////////////////////////////////////////
  // ini input parameter
  //////////////////////////////////////////////////

  // please set the variable LicenceCode to the value offered from Vision Components
  PatPar->LicenceCode1   = 0x1234ABCD;
  PatPar->LicenceCode2   = 0xABCD1234;
  

  // AngleMin [-360..360] and AngleMax [AngleMin..AngleMin+360] (for example -45 till +45 degrees: AngleMin=-45 AngleMax=45). 
  PatPar->AngleMin      = -180;
  PatPar->AngleMax      =  180;
  PatPar->AngleStep     =    5;
  

  // Scaling Range in percent [-50%..50%] 
  PatPar->ScaleMin      =   -0; // -40
  PatPar->ScaleMax      =    0; // +40
  PatPar->ScaleStep     =   10;
  

  //////////////////////////////////////////////////
  // speed and safity parameter
  //////////////////////////////////////////////////

  // pyramid zoom for search area (speed improvements)
  PatPar->PyrZoom       =   0;

  // SearchAcc[] (0 = Low / 1 = High Subpixel and Pattern Accuracy)
  // [0]: Search Step
  // [1]: Adjust Step
  PatPar->SearchAcc[0]  =   0; 
  PatPar->SearchAcc[1]  =   1; 

  // SpeedLevel: [0 = Safe .. 9 = Fast]
  PatPar->SpeedLevel    =   5;

  
  //////////////////////////////////////////////////
  // expert search parameter
  //////////////////////////////////////////////////

  // Expert Mode (0 = Overwrite Expert Parameter depending on SpeedLevel / 1 = Use Expert Parameter)
  PatPar->ExpertMode    =   0;

  // edge detection parameter for VC function edge()
  PatPar->Type          =   4;
  PatPar->Sigma         =  58;
  PatPar->BinMode       =   2;
  PatPar->MinContr      =   5;
  PatPar->Thresh        =  20;


  // Define Pattern Anchor Point Type
  // Typ  1: Center of Pattern Bounding Box (recommended)
  // Typ  0: Use defined AnchorX and AnchorY Points inside Pattern Area (experienced user only)
  PatPar->AnchorTyp     =   1;

  // set anchor point, necessary if AnchorTyp is 0
  PatPar->AnchorX       =  32; 
  PatPar->AnchorY       =  32;

  // search parameter
  PatPar->PatRes[0]     =   4; 
  PatPar->SrcRes[0]     =   6; 
  PatPar->PatRes[1]     =   0; 
  PatPar->SrcRes[1]     =   0; 

  // Memory Level [1 = high .. 3 = low memory consumption]
  PatPar->MemLevel      =   2;

  // Memory Mode (0 = no / 1 = low / 2 = high cache optimization)
  PatPar->MemMode       =   1; 
  

  //////////////////////////////////////////////////
  // parameter set ready 
  //////////////////////////////////////////////////
  
  return 0;
}

/*************************************************************************************************************************/

void ChangeParameter (PatternPar *PatPar)
{
  I32   response;

  EmptyKeyboardBufferAll();

  do
  {
    print("\n"); 
    print("--------------------------------------------------------------------------------\n");
    print("p:   change search parameter\n"); 
    print("a:   Change angle parameter.\n"); 
    print("s:   Change scale parameter.\n");
    print("e:   Change expert parameter.\n");
    print("\n"); 
    print("q:   quit Menu\n"); 
    print("--------------------------------------------------------------------------------\n");

    response = ReadKeyboard(1);
  // SearchAcc[] (0 = Low / 1 = High Subpixel and Pattern Accuracy)
  // [0]: Search Step
  // [1]: Adjust Step

    switch(response)
    {
      case 'p' : 
                 PatPar->ExpertMode = 0;
                 ChangeValue(&PatPar->PyrZoom,       0,   2,   1, "Search Image Pyramid");
                 ChangeValue(&PatPar->SearchAcc[0],  0,   1,   1, "Accuracy Search Step (1=High / 0=Low)");
                 ChangeValue(&PatPar->SearchAcc[1],  0,   1,   1, "Accuracy Adjust Step (1=High / 0=Low)");
                 ChangeValue(&PatPar->SpeedLevel,    0,   9,   1, "Speed Level [0=Safe .. 9=Fast]");
                 break;

      case 'a' : 
                 ChangeValue(&PatPar->AngleMin,  -360,  360,  5, "Min Rotation Angle in degrees");
                 if(PatPar->AngleMax < PatPar->AngleMin) PatPar->AngleMax = PatPar->AngleMin;
                 ChangeValue(&PatPar->AngleMax,   PatPar->AngleMin, PatPar->AngleMin + 360, 5, "Max Rotation Angle in degrees");
                 ChangeValue(&PatPar->AngleStep,     1,  10, 1, "Rotation Step Size in degrees");
                 break;

      case 's' :  
                 ChangeValue(&PatPar->ScaleMin,   -50,    0,   5, "Min Scaling Factor in %"); 
                 ChangeValue(&PatPar->ScaleMax,     0,   50,   5, "Max Scaling Factor in %");
                 ChangeValue(&PatPar->ScaleStep,    1,   20,   1, "Step Scale in %"); 
                 break;
                 
      case 'e' : 
                 PatPar->ExpertMode = 1;
                 ChangeValue(&PatPar->PyrZoom,       0,   2,   1, "Search Image Pyramid");
                 ChangeValue(&PatPar->SearchAcc[0],  0,   1,   1, "Accuracy Search Step (1=High / 0=Low)");
                 ChangeValue(&PatPar->SearchAcc[1],  0,   1,   1, "Accuracy Adjust Step (1=High / 0=Low)");
                 ChangeValue(&PatPar->MemMode,       0,   2,   1, "Memory Mode (Cache Optimization)");
                 ChangeValue(&PatPar->MemLevel,      1,   3,   1, "Memory Level");
                 ChangeValue(&PatPar->PatRes[0],     0,  10,   2, "Pattern Resolution for Search");
                 ChangeValue(&PatPar->SrcRes[0],     0,  10,   1, "Source Resolution for Search");
                 ChangeValue(&PatPar->PatRes[1],     0,  10,   2, "Pattern Resolution for Adjust");
                 ChangeValue(&PatPar->SrcRes[1],     0,  10,   1, "Source Resolution for Adjust");
                 break;

    }
  }
  while(response != 'q');
  

  /////////////////////////////////////////////////////////////////////////////////////
  // always call TeachParameterChange() functions after parameter changes 
  /////////////////////////////////////////////////////////////////////////////////////
  
  TeachParameterChange(PatPar);


  print("RUN mode\n"); 
}

/************************************************************************************************************************/

I32 TeachPattern (PatternPar *PatPar, image *Pat) 
{
  I32   i, j, PatDx, PatDy, response, Thresh, Error;
  image ImgEdge, Ovl;
  I32   BlobSize, MaxBlob, MaxNr, MaxRLC, MinObjSize, MaxObj, NrObj;
  U16   *rlc;
  ftr   *f;

  /////////////////////////////////////////////////////////////////////////////////////
  // ini valus
  /////////////////////////////////////////////////////////////////////////////////////

  Error         = 0;
  Thresh        = 1;
  PatDx         = Pat->dx;
  PatDy         = Pat->dy;

  ImageAssign(&Ovl, OvlByteAddr(ScrGetX(Pat->st), ScrGetY(Pat->st)), Pat->dx, Pat->dy, OvlGetPitch);

 
  /////////////////////////////////////////////////////////////////////////////////////
  // build edge image
  /////////////////////////////////////////////////////////////////////////////////////

  tpict();
  
  ImageAllocate(&ImgEdge, IMAGE_VECTOR, PatDx, PatDy);
  if (ImgEdge.st == NULL) {print("Error: Edge Image Allocation\n"); goto END;}
  
  edge(Pat, &ImgEdge, PatPar->Type, (float)PatPar->Sigma / 10.0, PatPar->BinMode, PatPar->MinContr, (float)PatPar->Thresh / 100.0, 128);

  
  /////////////////////////////////////////////////////////////////////////////////////
  // ini rlc values
  /////////////////////////////////////////////////////////////////////////////////////
  
  MaxObj        = 40000;
  MinObjSize    =     6;
  MaxRLC        = (PatDx * PatDy);

  rlc           = NULL;
  f             = NULL;
  

  /////////////////////////////////////////////////////////////////////////////////////
  // allocate RLC and feature memory
  /////////////////////////////////////////////////////////////////////////////////////

  rlc    = rlcmalloc(MaxRLC);
  if (rlc    == NULL) {Error = -1; goto END;}

  f      = (ftr *)sysmalloc((MaxObj * sizeof(ftr) - 1) / sizeof(I32) + 1, MDATA);
  if (f      == NULL) {Error = -2; goto END;}

  
  /////////////////////////////////////////////////////////////////////////////////////
  // do RLC coding and blob analyse
  /////////////////////////////////////////////////////////////////////////////////////

  NrObj = FindBlobs(&ImgEdge, rlc, MaxRLC, Thresh, MinObjSize, f, MaxObj);    


  /////////////////////////////////////////////////////////////////////////////////////
  // select edges
  /////////////////////////////////////////////////////////////////////////////////////

  for (i=0; i<NrObj; i++)
  {
    if (f[i].color)
    {
      DisplayBlob(&Ovl, rlc, i, COLBLUE);
    }
  }

  // delete old edges
  set(&ImgEdge, 0);
  

  /////////////////////////////////////////////////////////////////////////////////////
  // select pattern
  /////////////////////////////////////////////////////////////////////////////////////

  print("\nPlease teach Pattern. Define Pattern Edges:\n\n");
  print("Press ENTER to use all edges\n");
  print("Press  'F2' to select edges and ENTER when ready.\n");

  response = ReadKeyboard(1);
  
  if(response == KEY_F2)
  {
    /////////////////////////////////////////////////////////////////////////////////////
    // select edges
    /////////////////////////////////////////////////////////////////////////////////////

    for (j=0; j<NrObj; j++)
    {
      // looking for the largest blob
      MaxBlob = 0;
      MaxNr   = 0;

      for (i=0; i<NrObj; i++)
      {
        if (f[i].color)
        {
          BlobSize = f[i].area;

          if(BlobSize > MaxBlob) {MaxBlob = BlobSize; MaxNr = i;}
        }
      }

      // don't look for this blob again
      f[MaxNr].area = 0;
      
      // check blob
      if (MaxBlob)
      {
        DisplayBlob(&Ovl, rlc, MaxNr, COLGREEN);
        
        EmptyKeyboardBufferAll();
        response = ReadKeyboard(1);

        if(response == KEY_F2)
        {
          DisplayBlob(&Ovl,     rlc, MaxNr, COLRED);
          DisplayBlob(&ImgEdge, rlc, MaxNr, 255);
        }
        else
        {
          DisplayBlob(&Ovl, rlc, MaxNr, 0);
        }
      
        if(response == KEY_ENTER) break;
      }
    }
    
    // delete not using blobs (extension lib function !)
    // rlc_delete(U16 *Pat, U16 *dst, select);
  }
  else
  {
    // take all edges
    for (i=0; i<NrObj; i++)
    {
      if (f[i].color)
      {
        DisplayBlob(&ImgEdge, rlc, i, 255);
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////////////
  // release memory
  /////////////////////////////////////////////////////////////////////////////////////
  
  if(f            != NULL) sysfree(f);
  if(rlc          != NULL) rlcfree(rlc);


  /////////////////////////////////////////////////////////////////////////////////////
  // teach pattern
  /////////////////////////////////////////////////////////////////////////////////////
  
  Error = TeachPatternEdge(PatPar, &ImgEdge);
  if(Error) {goto END;}


  /////////////////////////////////////////////////////////////////////////////////////
  // draw selected pattern
  /////////////////////////////////////////////////////////////////////////////////////

  set(&Ovl, 0);
  Error = DrawPattern(PatPar, &Ovl, COLRED);
  // MarkCross  (&Ovl, Round(PatPar->PosX), Round(PatPar->PosY), COLRED, 10);


  /////////////////////////////////////////////////////////////////////////////////////
  // free memory
  /////////////////////////////////////////////////////////////////////////////////////

  ImageFree(&ImgEdge);

  print("\nPattern Teach Ready.\n");

  END:


  return Error;
}

/************************************************************************************************************************/

I32 FindBlobs (image *Src, U16 *rlc, I32 MaxRLC, I32 Thresh, I32 DelMinArea, ftr *f, I32 MaxObjects)

{
  I32 objects, ConnectionMode = 2;
  U16 *slc;

  // create RLC
  slc = rlcmk(Src, Thresh, rlc, MaxRLC);
  if (slc == NULL) return(-1); // pstr("RLC overrun\n");


  // delete small objects
  if (DelMinArea)
  {
    if(rlc_label(rlc, slc, ConnectionMode) == NULL) return(-2); // pstr("label 2 overrun\n");
    rlc_arf(rlc, rlc, DelMinArea);
  }
  
  // object labelling
  if(rlc_label(rlc, slc, ConnectionMode) == NULL) return(-3); // pstr("label 2 overrun\n");

  /* feature extraction */
  objects = rl_ftr2(rlc, f, MaxObjects);
  if(objects < 0) return(-4); // pstr("object number overrun\n");

  return(objects);
}

/************************************************************************************************************************/

I32 DisplayBlob(image *Area, U16 *src, I32 objnum, I32 color)
{
  U16 *slc;
  U8  *pPix;
  I32 dx, dy, x, y, obj, pos, ObjStart, ObjEnd, pitch;

  pPix  = (U8 *)Area->st;
  pitch =       Area->pitch;
  
  slc = *(U16 **)src;  			/* read slc address  			*/
  src += 2;

  // check input slc
  if(slc    == NULL) return(-1);

  // object number not existent
  if(objnum >= *slc) return(-2);

  // or object number overrun   	*/
  if(*slc    ==   0) return(-3);        			

  dx = *src++;           		/* read line delimiter          */
  dy = *src;            		/* read dy                      */

  // print("dx=%d dy=%d\n", dx, dy); 
  
  if (dx > Area->dx) return -4;
  if (dy > Area->dy) return -5;

  
  for(y=0; y<dy; y++)
  {
     slc++;                       /* skip dummy                   */
     src++;                       /* skip color                   */

     do
     {
       obj = *slc++;
       pos = *src++;

       // print("objnum=%3d line=%3d obj=%3d  s=%3d  e=%3d\n", objnum, y, obj, pos, *src); 

       if(obj == objnum)
       {
         ObjStart = pos;
         ObjEnd   = *src;
         
         for(x=ObjStart; x<ObjEnd; x++) *(pPix + x + y * pitch) = color;
       } 
     }
     while(*src != dx);

  }

  return 0;
}

/************************************************************************************************************************/

I32 StorePattern (PatternPar *PatPar, char *FileName)
{
  I32  res;
  FILE *fp;
  
  // delete old file if it exists 
  fp = io_fopen(FileName,"c");

  if(fp)
  {
    // print("delete old file %s\n",FileName);
    io_ioctl(fp, IO_DEL, NULL);
    io_fclose(fp);
  }
  else
  {
    // print("no old file to delete\n");
  }


  // write file 
  fp = io_fopen(FileName,"w");

  if(fp)
  {
    res = io_write(fp, (char *)PatPar->PatternData, PatPar->PatternLen);
    if (res >= 0)  print("Store Pattern (File %s  PatLen=%d Bytes)\n", FileName, res);
    else           print("can't write to file %s\n", FileName);
    io_fclose(fp);
  }
  else
  {
    print("cannot open file for write\n");
  }
  
  return(res);
}

/************************************************************************************************************************/

I32 ReadPattern (PatternPar *PatPar, char *FileName)
{
  I32  res, PatLen;
  U8   *PatDat;
  FILE *fp;
  
  // read data from file

  fp = io_fopen(FileName,"r");;

  if(fp)
  {
    PatLen = fp->SIZE;
    
    if (PatLen)
    {
      // allocate memory for pattern
      PatDat = (U8  *)sysmalloc((PatLen * sizeof(U8 ) - 1) / sizeof(I32) + 1, MDATA);

      if (PatDat == NULL) 
      {
        print("memory allocation error\n");
      }
      else
      {
        res = io_read(fp, (char *)PatDat, PatLen);

        if (res == PatLen)
        {
          res = TeachPatternData(PatPar, PatDat, PatLen);
          
          if (res == 0) print("Read Pattern From File %s  (PatLen=%d)\n", FileName, PatLen);
          else          print("Read Pattern From File Error %d\n", res);
        }
        else
        {
          print("can't read from file %s\n", FileName);
        }
        
        sysfree(PatDat);
      }
    }

    io_fclose(fp);
  }
  else
  {
    print("cannot open file for read\n");
  }
  
  return(res);
}

/***************************************************************************************************************************/

void EmptyKeyboardBufferAll (void)
{
  while(kbready()) kbdrcv();
  while(rbready()) rs232rcv();
}

/*************************************************************************************************************************/

void SelectArea (I32 *px, I32 *py, I32 *pdx, I32 *pdy, I32 xmin, I32 xmax, I32 ymin, I32 ymax, I32 text, I32 even, I32 color)

  {
  I32   x, y, dx, dy, response, increment=10, mode=0, smallstep, bigstep;
  image Ovl, Text;
  char  textstr[40];

  print("\n");
  print("Press ENTER  to accept values\n");
  print("Press F1     step size\n");
  print("Press F2     toggle between Position and Size\n");
  print("Press CURSOR to change values\n");

  x  = *px;
  y  = *py;
  dx = *pdx;
  dy = *pdy;

  bigstep = 10;
  increment = bigstep;
  if(even) smallstep = 2;
  else     smallstep = 1;

  vmode(vmOvlLive);

  ImageAssign(&Ovl, OvlByteAddr(x, y), dx, dy, OvlGetPitch);  

  /* select size */
  do
  {
    framed(&Ovl, color);

    if(text)
    {
      ImageAssign(&Text, OvlBitAddr(x+2, y+2), dx, 8, OvlGetPitch);  
      sprint(textstr,"X=%d Y=%d DX=%d DY=%d", x, y, dx, dy); 
      chprint1(textstr,&Text, 1, 1, color);
    }

    response = ReadKeyboard(1);
    switch(response)
        {
        case KEY_DOWN : if(mode) {dy+=increment; if(y+dy > ymax) dy=ymax- y;}
                        else     { y+=increment; if(y+dy > ymax)  y=ymax-dy;} break;
        case KEY_RIGHT: if(mode) {dx+=increment; if(x+dx > xmax) dx=xmax- x;}
                        else     { x+=increment; if(x+dx > xmax)  x=xmax-dx;} break;
        case KEY_UP   : if(mode) {dy-=increment; if(dy <    2) dy=2;} 
                        else     { y-=increment; if(y  < ymin)  y=ymin;} break;
        case KEY_LEFT : if(mode) {dx-=increment; if(dx <    2) dx=2;} 
                        else     { x-=increment; if(x  < xmin)  x=xmin;} break;

        case KEY_F1   : if(increment==bigstep) increment=smallstep; else increment=bigstep; break;
        case KEY_F2   : if(mode     == 1) mode =0; else mode = 1; break;
        }

    if (even)
    {
      x  = EVEN(x);
      y  = EVEN(y);
      dx = EVEN(dx);
      dy = EVEN(dy);
    }

    if(text)
    {
      chprint1(textstr,&Text,1,1,0);
    }

    framed(&Ovl, 0);
    ImageAssign(&Ovl,OvlBitAddr(x,y), dx, dy, OvlGetPitch);  
  }
  while(response!=KEY_ENTER && response!=KEY_ESC);

  if (response==KEY_ENTER)
    {
    *px  =  x;
    *py  =  y;
    *pdx = dx;
    *pdy = dy;
    }
  }

/***************************************************************************************************************************/

void CapturePosition(I32 x, I32 y)
{
  image area;


  /* define image variable at the image page */
  ImageAssign(&area, ScrGetCaptPage, ScrGetColumns, ScrGetRows, ScrGetPitch);

  if (x < 0) x=0;
  if (x > DispGetColumns-ScrGetColumns) x=DispGetColumns-ScrGetColumns;

  x=(x/32)*32;

  if (y < 0) y=0;
  if (y > DispGetRows-ScrGetRows) y=DispGetRows-ScrGetRows;

  ScrSetCaptPage(ScrGetDispPage + x + y*ScrGetPitch);
   
  /* delete drawings from old position */
  set(&area, 0);

  tpict(); // get new image
}

/*************************************************************************************************************************/

void ChangeColor (image *area, I32 start, I32 end, I32 newcol)

{
  I32 i, j, dx, dy, pitch;
  U8 *restrict ppix, *addr;

  
  addr  = area->st;
  dx    = area->dx;
  dy    = area->dy;
  pitch = area->pitch;

  for (j=0; j<dy; j++)
  {
    ppix = addr+(pitch*j);

    for (i=0; i<dx; i++)
    {
      if ( (*ppix>=start) && (*ppix<=end) )
      {
        *ppix=newcol;
      }
      ppix++;
    }
  }
}

/*************************************************************************************************************************/

void High (image *a)

{
  image area;
  I32   dx, dy;
  
  ImageAssign(&area, a->st-2-2*a->pitch, a->dx+4, a->dy+4, a->pitch);

  dx = area.dx-1;
  dy = area.dy-1;
  
  lined(&area,    0,    0,   dx,    0, COLWHITE);
  lined(&area,    1,    1, dx-1,    1, COLWHITE);
  lined(&area,    0,    0,    0,   dy, COLWHITE);
  lined(&area,    1,    1,    1, dy-1, COLWHITE);

  lined(&area,    1,   dy,   dx,   dy, COLBLACK);
  lined(&area,    2, dy-1, dx-1, dy-1, COLBLACK);
  lined(&area,   dx,    1,   dx,   dy, COLBLACK);
  lined(&area, dx-1,    2, dx-1, dy-1, COLBLACK);
}

/*************************************************************************************************************************/

void Low (image *a)

{
  image area;
  I32   dx, dy;
  
  ImageAssign(&area, a->st-2-2*a->pitch, a->dx+4, a->dy+4, a->pitch);
  
  dx = area.dx-1;
  dy = area.dy-1;

  lined(&area,    0,    0,   dx,    0, COLBLACK);
  lined(&area,    1,    1, dx-1,    1, COLBLACK);
  lined(&area,    0,    0,    0,   dy, COLBLACK);
  lined(&area,    1,    1,    1, dy-1, COLBLACK);

  lined(&area,    1,   dy,   dx,   dy, COLWHITE);
  lined(&area,    2, dy-1, dx-1, dy-1, COLWHITE);
  lined(&area,   dx,    1,   dx,   dy, COLWHITE);
  lined(&area, dx-1,    2, dx-1, dy-1, COLWHITE);
}

/***************************************************************************************************************************/

